# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import unittest

from lsst.daf.relation import (
    BinaryOperationRelation,
    ColumnError,
    ColumnExpression,
    EngineError,
    Join,
    Predicate,
    SortTerm,
    iteration,
    tests,
)


class JoinTestCase(tests.RelationTestCase):
    """Tests for the Join operation and relations based on it."""

    def setUp(self) -> None:
        self.a = tests.ColumnTag("a")
        self.b = tests.ColumnTag("b")
        self.c = tests.ColumnTag("c")
        self.engine = iteration.Engine(name="preferred")
        self.leaf_1 = self.engine.make_leaf(
            {self.a, self.b},
            payload=iteration.RowSequence(
                [{self.a: 0, self.b: 5}, {self.a: 1, self.b: 10}, {self.a: 2, self.b: 25}]
            ),
            name="leaf_1",
        )
        self.leaf_2 = self.engine.make_leaf(
            {self.a, self.c},
            payload=iteration.RowSequence(
                [{self.a: 0, self.c: 15}, {self.a: 2, self.c: 20}, {self.a: 3, self.b: 0}]
            ),
            name="leaf_2",
        )

    def test_attributes(self) -> None:
        """Check that all Relation and PartialJoin attributes have the expected
        values.
        """
        relation = self.leaf_1.join(self.leaf_2)
        assert isinstance(relation, BinaryOperationRelation)
        self.assertEqual(relation.columns, {self.a, self.b, self.c})
        self.assertEqual(relation.engine, self.engine)
        self.assertEqual(relation.min_rows, 0)
        self.assertEqual(relation.max_rows, 9)
        self.assertFalse(relation.is_locked)
        operation = relation.operation
        assert isinstance(operation, Join)
        self.assertEqual(operation.min_columns, {self.a})
        self.assertEqual(operation.max_columns, {self.a})
        self.assertEqual(operation.common_columns, {self.a})
        self.assertEqual(operation.predicate, Predicate.literal(True))
        partial = Join().partial(self.leaf_1)
        self.assertEqual(partial.columns_required, frozenset())
        self.assert_relations_equal(partial.fixed, self.leaf_1)
        self.assertFalse(partial.is_count_dependent)
        self.assertFalse(partial.is_order_dependent)
        self.assertFalse(partial.is_count_invariant)
        self.assertFalse(partial.is_empty_invariant)
        self.assertEqual(partial.applied_columns(self.leaf_2), {self.a, self.b, self.c})
        self.assertEqual(partial.applied_min_rows(self.leaf_2), 0)
        self.assertEqual(partial.applied_max_rows(self.leaf_2), 9)

    def test_apply_failures(self) -> None:
        """Test failure modes of constructing and applying Join."""
        # Mismatched engines.
        new_engine = iteration.Engine(name="downstream")
        with self.assertRaises(EngineError):
            Join().apply(self.leaf_1.transferred_to(new_engine), self.leaf_2)
        # Predicate requires nonexistent columns.
        predicate = ColumnExpression.reference(tests.ColumnTag("d")).lt(ColumnExpression.literal(0))
        with self.assertRaises(ColumnError):
            Join(predicate=predicate).apply(self.leaf_1, self.leaf_2)
        with self.assertRaises(ColumnError):
            Join(predicate=predicate).partial(self.leaf_1).apply(self.leaf_2)
        with self.assertRaises(ColumnError):
            Join(predicate=predicate).partial(self.leaf_2).apply(self.leaf_1)
        # Bounds on columns internally inconsistent.
        with self.assertRaises(ColumnError):
            Join(min_columns=frozenset({self.a, self.b}), max_columns=frozenset({self.a}))
        # Minimum columns not satisfied.
        join = Join(min_columns=frozenset({self.a, self.b}))
        with self.assertRaises(ColumnError):
            join.apply(self.leaf_1, self.leaf_2)
        with self.assertRaises(ColumnError):
            join.apply(self.leaf_2, self.leaf_1)
        with self.assertRaises(ColumnError):
            join.partial(self.leaf_2)
        with self.assertRaises(ColumnError):
            join.partial(self.leaf_1).apply(self.leaf_2)
        # Common columns not satisfied.
        join = Join(min_columns=frozenset({self.a, self.b}), max_columns=frozenset({self.a, self.b}))
        with self.assertRaises(ColumnError):
            join.apply(self.leaf_1, self.leaf_2)
        with self.assertRaises(ColumnError):
            join.apply(self.leaf_2, self.leaf_1)

    def test_apply_simplify(self) -> None:
        """Test Join.apply simplifications."""
        join_identity = self.engine.make_join_identity_relation()
        self.assertIs(self.leaf_1.join(join_identity), self.leaf_1)
        self.assertIs(join_identity.join(self.leaf_1), self.leaf_1)

    def test_backtracking_apply(self) -> None:
        """Test `PartialJoin.apply` logic that involves reordering operations
        in the existing tree to perform the new operation in a preferred
        engine.
        """
        new_engine = iteration.Engine(name="downstream")
        d = tests.ColumnTag("d")
        expression = ColumnExpression.function(
            "__add__", ColumnExpression.reference(self.a), ColumnExpression.literal(5)
        )
        sort_terms = [SortTerm(ColumnExpression.reference(self.a))]
        predicate = ColumnExpression.reference(self.b).gt(ColumnExpression.literal(0))
        # Apply a bunch of operations in a new engine that a PartialJoin should
        # commute with.
        target = (
            self.leaf_1.transferred_to(new_engine)
            .with_calculated_column(d, expression)
            .with_rows_satisfying(predicate)
            .with_only_columns({self.a, d})
            .sorted(sort_terms)
        )
        # Apply a new PartialJoin with backtracking and see that it appears
        # before the transfer to the new engine, with adjustments as needed.
        relation = target.join(self.leaf_2)
        self.assert_relations_equal(
            relation,
            (
                self.leaf_1.join(self.leaf_2)
                .transferred_to(new_engine)
                .with_calculated_column(d, expression)
                .with_rows_satisfying(predicate)
                .with_only_columns({self.a, self.c, d})
                .sorted(sort_terms)
            ),
        )

    def test_no_backtracking(self) -> None:
        """Test `PartialJoin.apply` logic that handles differing engines
        without reordering operations in the existing tree, as well as failures
        in that reordering.
        """
        new_engine = iteration.Engine(name="downstream")
        # Construct a relation tree we can't reorder when inserting a Join,
        # because there is a locked Materialization in the way.
        target = self.leaf_1.transferred_to(new_engine).materialized("lock")
        # We can automatically transfer (back) to the new relation's engine.
        self.assert_relations_equal(
            target.join(self.leaf_2, transfer=True),
            target.transferred_to(self.engine).join(self.leaf_2),
        )
        # Can't backtrack through a Deduplication.
        target = self.leaf_1.transferred_to(new_engine).without_duplicates()
        with self.assertRaises(EngineError):
            target.join(self.leaf_2)
        # Can't backtrack through a Slice, because it's order/count dependent.
        target = self.leaf_1.transferred_to(new_engine)[:2]
        with self.assertRaises(EngineError):
            target.join(self.leaf_2)

    def test_common_columns(self) -> None:
        """Test Join.applied_common_columns logic."""
        leaf_3 = self.engine.make_leaf(
            {self.a, self.b, self.c},
            payload=iteration.RowSequence(
                [{self.a: 0, self.b: 2, self.c: 15}, {self.a: 2, self.b: 4, self.c: 20}]
            ),
            name="leaf_2",
        )
        # With no min or max columns, common_columns is just the intersection
        # of the columns of the operands.
        self.assertEqual(Join().applied_common_columns(self.leaf_1, leaf_3), {self.a, self.b})
        # Check that max_columns is enforced.
        self.assertEqual(
            Join(max_columns=frozenset({self.a})).applied_common_columns(self.leaf_1, leaf_3), {self.a}
        )
        # Check that min_columns is enforced.
        with self.assertRaises(ColumnError):
            Join(min_columns=frozenset({self.c})).applied_common_columns(self.leaf_1, leaf_3)
        # Repeat last two checks with min_columns == max_columns.
        self.assertEqual(
            Join(min_columns=frozenset({self.a}), max_columns=frozenset({self.a})).applied_common_columns(
                self.leaf_1, leaf_3
            ),
            {self.a},
        )
        with self.assertRaises(ColumnError):
            Join(min_columns=frozenset({self.c}), max_columns=frozenset({self.c})).apply(self.leaf_1, leaf_3)

    def test_str(self) -> None:
        """Test str(Join), str(PartialJoin), and
        str(BinaryOperationRelation[Join]).
        """
        relation = self.leaf_1.join(self.leaf_2)
        self.assertEqual(str(relation), "leaf_1 ⋈ leaf_2")
        partial = Join().partial(self.leaf_1)
        self.assertEqual(str(partial), "⋈[leaf_1]")
        # Nested operations get parentheses, unless they're joins or leaves.
        leaf_3 = self.engine.make_leaf(
            {self.a, self.b},
            payload=iteration.RowSequence([{self.a: 3, self.b: 4}]),
            name="leaf_3",
        )
        self.assertEqual(str(relation.join(leaf_3)), "leaf_1 ⋈ leaf_2 ⋈ leaf_3")
        self.assertEqual(str(self.leaf_1.chain(leaf_3).join(self.leaf_2)), "(leaf_1 ∪ leaf_3) ⋈ leaf_2")


if __name__ == "__main__":
    unittest.main()



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = (
    "UnaryOperationRelation",
    "BinaryOperationRelation",
)

import dataclasses
from collections.abc import Set
from typing import TYPE_CHECKING, Literal, final

from ._leaf_relation import LeafRelation
from ._relation import BaseRelation, Relation

if TYPE_CHECKING:
    from ._binary_operation import BinaryOperation
    from ._columns import ColumnTag
    from ._engine import Engine
    from ._unary_operation import UnaryOperation


@dataclasses.dataclass(frozen=True)
class UnaryOperationRelation(BaseRelation):
    """A concrete `Relation` that represents the action of a `UnaryOperation`
    on a target `Relation`.

    `UnaryOperationRelation` instances must only be constructed via calls to
    `UnaryOperation.apply` or `Relation` convenience methods.  Direct calls to
    the constructor are not guaranteed to satisfy invariants imposed by the
    operations classes.
    """

    operation: UnaryOperation
    """The unary operation whose action this relation represents
    (`UnaryOperation`).
    """

    target: Relation
    """The target relation the operation acts upon (`Relation`).
    """

    columns: Set[ColumnTag] = dataclasses.field(repr=False, compare=False)
    """The columns in this relation (`~collections.abc.Set` [ `ColumnTag` ] ).
    """

    @property
    def payload(self) -> None:
        """The engine-specific contents of the relation.

        This is always `None` for binary operation relations.
        """
        return None

    @property
    def is_locked(self) -> Literal[False]:
        """Whether this relation and those upstream of it should be considered
        fixed by tree-manipulation algorithms (`bool`).
        """
        return False

    @property
    def engine(self) -> Engine:
        """The engine that is responsible for interpreting this relation
        (`Engine`).
        """
        return self.target.engine

    @property
    def min_rows(self) -> int:
        """The minimum number of rows this relation might have (`int`)."""
        return self.operation.applied_min_rows(self.target)

    @property
    def max_rows(self) -> int | None:
        """The maximum number of rows this relation might have (`int` or
        `None`).

        This is `None` for relations whose size is not bounded from above.
        """
        return self.operation.applied_max_rows(self.target)

    def __str__(self) -> str:
        return f"{self.operation!s}({self.target!s})"

    def reapply(self, target: Relation) -> Relation:
        """Reapply this relation's operation with a possibly-new target.

        Parameters
        ----------
        target : `Relation`
            Possibly-new target.

        Returns
        -------
        relation : `Relation`
            Relation that applies `operation` to the new target.  Will be
            ``self`` if ``target is self.target`` (this avoidance of an
            unnecessary new `UnaryOperationRelation` instance is the main
            reason for this convenience method's existence).
        """
        if target is self.target:
            return self
        else:
            return self.operation.apply(target)


@final
@dataclasses.dataclass(frozen=True)
class BinaryOperationRelation(BaseRelation):
    """A concrete `Relation` that represents the action of a `BinaryOperation`
    on a pair of target `Relation` objects.

    `BinaryOperationRelation` instances must only be constructed via calls to
    `BinaryOperation.apply` or `Relation` convenience methods.  Direct calls to
    the constructor are not guaranteed to satisfy invariants imposed by the
    operations classes.
    """

    operation: BinaryOperation
    """The binary operation whose action this relation represents
    (`BinaryOperation`).
    """

    lhs: Relation
    """One target relation the operation acts upon (`Relation`).
    """

    rhs: Relation
    """The other target relation the operation acts upon (`Relation`).
    """

    columns: Set[ColumnTag] = dataclasses.field(repr=False, compare=False)
    """The columns in this relation (`~collections.abc.Set` [ `ColumnTag` ] ).
    """

    @property
    def is_locked(self) -> Literal[False]:
        """Whether this relation and those upstream of it should be considered
        fixed by tree-manipulation algorithms (`bool`).
        """
        return False

    @property
    def engine(self) -> Engine:
        """The engine that is responsible for interpreting this relation
        (`Engine`).
        """
        return self.lhs.engine

    @property
    def payload(self) -> None:
        """The engine-specific contents of the relation.

        This is always `None` for binary operation relations.
        """
        return None

    @property
    def min_rows(self) -> int:
        """The minimum number of rows this relation might have (`int`)."""
        return self.operation.applied_min_rows(self.lhs, self.rhs)

    @property
    def max_rows(self) -> int | None:
        """The maximum number of rows this relation might have (`int` or
        `None`).

        This is `None` for relations whose size is not bounded from above.
        """
        return self.operation.applied_max_rows(self.lhs, self.rhs)

    def __str__(self) -> str:
        lhs_str = f"({self.lhs!s})"
        match self.lhs:
            case LeafRelation():
                lhs_str = str(self.lhs)
            case BinaryOperationRelation(operation=lhs_operation):
                if type(lhs_operation) is type(self.operation):  # noqa: E721
                    lhs_str = str(self.lhs)
        rhs_str = f"({self.rhs!s})"
        match self.rhs:
            case LeafRelation():
                rhs_str = str(self.rhs)
            case BinaryOperationRelation(operation=rhs_operation):
                if type(rhs_operation) is type(self.operation):  # noqa: E721
                    rhs_str = str(self.rhs)
        return f"{lhs_str} {self.operation!s} {rhs_str}"

    def reapply(self, lhs: Relation, rhs: Relation) -> Relation:
        """Reapply this relation's operation with possibly-new targets.

        Parameters
        ----------
        lhs : `Relation`
            One possibly-new target.
        rhs : `Relation`
            The other possibly-new target.

        Returns
        -------
        relation : `Relation`
            Relation that applies `operation` to the new targets.  Will be
            ``self`` if ``lhs is self.lhs and rhs is self.rhs`` (this avoidance
            of an unnecessary new `BinaryOperationRelation` instance is the
            main reason for this convenience method's existence).
        """
        if lhs is self.lhs and rhs is self.rhs:
            return self
        else:
            return self.operation.apply(lhs, rhs)



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("BinaryOperation", "IgnoreOne")

import dataclasses
from abc import ABC, abstractmethod
from collections.abc import Set
from typing import final

from ._columns import ColumnTag
from ._relation import Relation


class BinaryOperation(ABC):
    """An abstract base class for operations that act on a pair of relations.

    Notes
    -----
    A `BinaryOperation` represents the operation itself; the combination of an
    operation and the "lhs" and "rhs" relations it acts on to form a new
    relation is represented by the `BinaryOperationRelation` class, which
    should always be performed via a call to the `apply` method (or something
    that calls it, like the convenience methods on the `Relation` class).  In
    many cases, applying a `BinaryOperation` doesn't return something involving
    the original operation, because of some combination of defaulted-parameter
    population and simplification, and there are even some `BinaryOperation`
    classes that should never actually appear in a `BinaryOperationRelation`.

    `BinaryOperation` cannot be subclassed by external code.

    All concrete `BinaryOperation` types are frozen, equality-comparable
    `dataclasses`.  They also provide a very concise `str` representation (in
    addition to the dataclass-provided `repr`) suitable for summarizing an
    entire relation tree.

    See Also
    --------
    :ref:`lsst.daf.relation-overview-operations`
    """

    def __init_subclass__(cls) -> None:
        assert cls.__name__ in {
            "Join",
            "Chain",
            "IgnoreOne",
        }, "BinaryOperation inheritance is closed to predefined types in daf_relation."

    @final
    def apply(self, lhs: Relation, rhs: Relation) -> Relation:
        """Create a new relation that represents the action of this operation
        on a pair of existing relations.

        Parameters
        ----------
        lhs : `Relation`
            One relation the operation will act on.
        rhs : `Relation`
            The other relation the operation will act on.

        Returns
        -------
        new_relation : `Relation`
            Relation that includes this operation.  This may be ``self`` if the
            operation is a no-op, and it may not be a `BinaryOperationRelation`
            holding this operation (or even a similar one) if the operation was
            inserted earlier in the tree via commutation relations.

        Raises
        ------
        ColumnError
            Raised if the operation could not be applied due to problems with
            the target relations' columns.
        EngineError
            Raised if the operation could not be applied due to problems with
            the target relations' engine(s).
        """
        operation = self._begin_apply(lhs, rhs)
        return lhs.engine.append_binary(operation, lhs, rhs)

    def _begin_apply(self, lhs: Relation, rhs: Relation) -> BinaryOperation:
        """A customization hook for the beginning of operation application.

        Parameters
        ----------
        lhs : `Relation`
            One relation the operation should act on.
        rhs : `Relation`
            The other relation the operation should act on.

        Returns
        -------
        operation : `BinaryOperation`
            The operation to actually apply.  The default implementation
            returns ``self``.

        Notes
        -----
        This method provides an opportunity for operations to establish any
        invariants that must be satisfied only when the operation is part of
        a relation.
        """  # noqa: D401
        return self

    def _finish_apply(self, lhs: Relation, rhs: Relation) -> Relation:
        """A customization hook for the end of operation application.

        Parameters
        ----------
        target : `Relation`
            Relation the operation will act upon directly.

        Returns
        -------
        applied : `Relation`
            Result of applying this operation to the given target.  Usually -
            but not always - a `UnaryOperationRelation` that holds ``self`` and
            ``target``.

        Notes
        -----
        This method provides an opportunity for operations to change the kind
        of relation produced (the default implementation constructs a
        `BinaryOperationRelation`).
        """  # noqa: D401
        from ._operation_relations import BinaryOperationRelation

        return BinaryOperationRelation(
            operation=self,
            lhs=lhs,
            rhs=rhs,
            columns=self.applied_columns(lhs, rhs),
        )

    @abstractmethod
    def applied_columns(self, lhs: Relation, rhs: Relation) -> Set[ColumnTag]:
        """Return the columns of the relation that results from applying this
        operation to the given targets.

        Parameters
        ----------
        lhs : `Relation`
            On relation the operation will act on.
        rhs : `Relation`
            The other relation the operation will act on.

        Returns
        -------
        columns : `~collections.abc.Set` [ `ColumnTag` ]
            Columns the new relation would have.
        """
        raise NotImplementedError()

    @abstractmethod
    def applied_min_rows(self, lhs: Relation, rhs: Relation) -> int:
        """Return the minimum number of rows of the relation that results from
        applying this operation to the given targets.

        Parameters
        ----------
        lhs : `Relation`
            On relation the operation will act on.
        rhs : `Relation`
            The other relation the operation will act on.

        Returns
        -------
        min_rows : `int`
            Minimum number of rows the new relation would have.
        """
        raise NotImplementedError()

    @abstractmethod
    def applied_max_rows(self, lhs: Relation, rhs: Relation) -> int | None:
        """Return the maximum number of rows of the relation that results from
        applying this operation to the given target.

        Parameters
        ----------
        lhs : `Relation`
            On relation the operation will act on.
        rhs : `Relation`
            The other relation the operation will act on.

        Returns
        -------
        max_rows : `int` or `None`
            Maximum number of rows the new relation would have.
        """
        raise NotImplementedError()


@dataclasses.dataclass
class IgnoreOne(BinaryOperation):
    """A binary operation that passes through one of its operands and ignores
    the other.
    """

    ignore_lhs: bool
    """Whether the ignored operand is the left-hand-side one (`bool`).
    """

    def applied_columns(self, lhs: Relation, rhs: Relation) -> Set[ColumnTag]:
        # Docstring inherited.
        return self._finish_apply(lhs, rhs).columns

    def applied_min_rows(self, lhs: Relation, rhs: Relation) -> int:
        # Docstring inherited.
        return self._finish_apply(lhs, rhs).min_rows

    def applied_max_rows(self, lhs: Relation, rhs: Relation) -> int | None:
        # Docstring inherited.
        return self._finish_apply(lhs, rhs).max_rows

    def _finish_apply(self, lhs: Relation, rhs: Relation) -> Relation:
        # Docstring inherited.
        if self.ignore_lhs:
            return rhs
        else:
            return lhs



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import unittest

from lsst.daf.relation import (
    ColumnExpression,
    EngineError,
    LeafRelation,
    Relation,
    Slice,
    UnaryOperationRelation,
    iteration,
    tests,
)


class SliceTestCase(tests.RelationTestCase):
    """Tests for the Slice operation and relations based on it."""

    def setUp(self) -> None:
        self.a = tests.ColumnTag("a")
        self.engine = iteration.Engine(name="preferred")
        self.leaf = self.engine.make_leaf(
            {self.a}, payload=iteration.RowSequence([{self.a: 0}, {self.a: 1}]), name="leaf"
        )

    def test_attributes(self) -> None:
        """Check that all UnaryOperation and Relation attributes have the
        expected values.
        """
        relation = self.leaf[1:2]
        assert isinstance(relation, UnaryOperationRelation)
        self.assertEqual(relation.columns, {self.a})
        self.assertEqual(relation.engine, self.engine)
        self.assertEqual(relation.min_rows, 1)
        self.assertEqual(relation.max_rows, 1)
        self.assertFalse(relation.is_locked)
        operation = relation.operation
        assert isinstance(operation, Slice)
        self.assertEqual(operation.start, 1)
        self.assertEqual(operation.stop, 2)
        self.assertEqual(operation.limit, 1)
        self.assertEqual(operation.columns_required, frozenset())
        self.assertFalse(operation.is_empty_invariant)
        self.assertFalse(operation.is_count_invariant)
        self.assertTrue(operation.is_order_dependent)
        self.assertTrue(operation.is_count_dependent)
        # Also check min/max attributes an unbounded Slice, since that involves
        # a few different logic branches.
        relation = self.leaf[1:]
        assert isinstance(relation, UnaryOperationRelation)
        self.assertEqual(relation.columns, {self.a})
        self.assertEqual(relation.engine, self.engine)
        self.assertEqual(relation.min_rows, 1)
        self.assertEqual(relation.max_rows, 1)
        self.assertFalse(relation.is_locked)
        operation = relation.operation
        assert isinstance(operation, Slice)
        self.assertEqual(operation.start, 1)
        self.assertEqual(operation.stop, None)
        self.assertEqual(operation.limit, None)
        self.assertEqual(operation.columns_required, frozenset())
        self.assertFalse(operation.is_empty_invariant)
        self.assertFalse(operation.is_count_invariant)
        self.assertTrue(operation.is_order_dependent)
        self.assertTrue(operation.is_count_dependent)

    def test_min_max_rows(self) -> None:
        """Test min_rows and max_rows for different kinds of slices
        and original min/max rows.
        """
        # iteration.Engine.make_leaf sets min_rows and max_rows based on
        # len(payload), which we don't want here.
        leaf1 = LeafRelation(self.engine, frozenset({self.a}), payload=..., min_rows=0, max_rows=None)
        leaf2 = LeafRelation(self.engine, frozenset({self.a}), payload=..., min_rows=0, max_rows=5)
        leaf3 = LeafRelation(self.engine, frozenset({self.a}), payload=..., min_rows=5, max_rows=5)
        leaf4 = LeafRelation(self.engine, frozenset({self.a}), payload=..., min_rows=5, max_rows=8)
        leaf5 = LeafRelation(self.engine, frozenset({self.a}), payload=..., min_rows=5, max_rows=None)

        # Reasoning about the expected values of slice operations is really
        # easy to get wrong, so instead we brute-force the expected values,
        # ultimately delegating to Python's own implementation of slicing
        # range objects.

        def brute_force_row_bounds(
            input_min_rows: int, input_max_rows: int | None, start: int, stop: int | None
        ) -> tuple[int, int | None]:
            """Compute the minimum and maximum number of rows a sequence could
            have after slicing.

            Parameters
            ----------
            input_min_rows, input_min_rows : `int` or `None`
                Original bounds on the number of rows.
            start, stop: `int` or `None`
                Slice parameters

            Returns
            -------
            output_min_rows, output_min_rows : `int` or `None`
                Bounds on the number of rows for the sliced sequence.

            Notes
            -----
            Since this is just a test helper, we handle `None` by assuming it
            can be replaced by a large value and that large values in the
            results indicate a `None` result.  Keep all concrete integers below
            100 to avoid problems.
            """
            sizes = []
            if input_max_rows is None:
                output_min_rows, output_max_rows = brute_force_row_bounds(input_min_rows, 100, start, stop)
                if output_max_rows is not None and output_max_rows > 50:
                    output_max_rows = None
                return output_min_rows, output_max_rows
            for n_rows in range(input_min_rows, input_max_rows + 1):
                sequence = range(n_rows)
                sizes.append(len(sequence[slice(start, stop)]))
            return min(sizes), max(sizes)

        def check(leaf: Relation) -> None:
            """Run tests on the given leaf relation by applying slices with
            a number of start and stop values that are just above, just below,
            or equal to its min and max rows.
            """
            breaks_set = {0, leaf.min_rows - 1, leaf.min_rows, leaf.min_rows + 1}
            if leaf.max_rows is not None:
                breaks_set.update({leaf.max_rows - 1, leaf.max_rows, leaf.max_rows + 1})
            breaks_list = list(breaks_set)
            breaks_list.sort()
            for start in breaks_list:
                for stop in breaks_list + [None]:
                    if start < 0:
                        with self.assertRaises(ValueError):
                            Slice(start, stop)
                    elif stop is not None and stop < start:
                        with self.assertRaises(ValueError):
                            Slice(start, stop)
                    else:
                        relation = leaf[slice(start, stop)]
                        self.assertEqual(
                            (relation.min_rows, relation.max_rows),
                            brute_force_row_bounds(leaf.min_rows, leaf.max_rows, start, stop),
                            msg=(
                                f"leaf.min_rows={leaf.min_rows}, "
                                f"leaf.max_rows={leaf.max_rows}, "
                                f"slice=[{start}:{stop}]"
                            ),
                        )

        check(leaf1)
        check(leaf2)
        check(leaf3)
        check(leaf4)
        check(leaf5)

    def test_backtracking_apply(self) -> None:
        """Test apply logic that involves reordering operations in the existing
        tree to perform the new operation in a preferred engine.
        """
        new_engine = iteration.Engine(name="downstream")
        b = tests.ColumnTag("b")
        expression = ColumnExpression.function(
            "__add__", ColumnExpression.reference(self.a), ColumnExpression.literal(5)
        )
        # Apply operations in a new engine that a Slice should commute with.
        target = (
            self.leaf.transferred_to(new_engine).with_calculated_column(b, expression).with_only_columns({b})
        )
        # Apply a new Slice with backtracking and see that it appears
        # before the transfer to the new engine.
        relation = Slice(start=1, stop=3).apply(
            target, preferred_engine=self.engine, require_preferred_engine=True
        )
        self.assert_relations_equal(
            relation,
            (
                self.leaf[1:3]
                .transferred_to(new_engine)
                .with_calculated_column(b, expression)
                .with_only_columns({b})
            ),
        )

    def test_no_backtracking(self) -> None:
        """Test apply logic that handles preferred engines without reordering
        operations in the existing tree.
        """
        new_engine = iteration.Engine(name="downstream")
        # Construct a relation tree we can't reorder when inserting a
        # Sort, because there is a locked Materialization in the way.
        target = self.leaf.transferred_to(new_engine).materialized("lock")
        # Preferred engine is ignored if we can't backtrack and don't enable
        # anything else.
        self.assert_relations_equal(
            Slice(start=1, stop=3).apply(target, preferred_engine=self.engine),
            Slice(start=1, stop=3).apply(target),
        )
        # We can force this to be an error.
        with self.assertRaises(EngineError):
            Slice(start=1, stop=3).apply(target, preferred_engine=self.engine, require_preferred_engine=True)
        # We can also automatically transfer (back) to the preferred engine.
        self.assert_relations_equal(
            Slice(start=1, stop=3).apply(target, preferred_engine=self.engine, transfer=True),
            target.transferred_to(self.engine)[1:3],
        )
        # Can't backtrack through anything other than a Projection or
        # a Calculation.
        target = self.leaf.transferred_to(new_engine).without_duplicates()
        with self.assertRaises(EngineError):
            Slice(start=1, stop=3).apply(target, preferred_engine=self.engine, require_preferred_engine=True)

    def test_apply_simplify(self) -> None:
        """Test simplification logic in Slice.apply."""
        # Test that applying a Slice to an existing Slice merges them.
        self.assert_relations_equal(self.leaf[1:][:2], self.leaf[1:3])
        self.assert_relations_equal(self.leaf[1:][1:], self.leaf[2:])
        self.assert_relations_equal(self.leaf[1:3][1:2], self.leaf[2:3])
        self.assert_relations_equal(self.leaf[1:3][1:], self.leaf[2:3])
        # Test that a no-op slice does nothing.
        self.assert_relations_equal(self.leaf[:], self.leaf)

    def test_iteration(self) -> None:
        """Test Slice execution in the iteration engine."""
        self.assertEqual(list(self.engine.execute(self.leaf[1:])), [{self.a: 1}])
        self.assertEqual(list(self.engine.execute(self.leaf[:1])), [{self.a: 0}])
        self.assertEqual(list(self.engine.execute(self.leaf[1:2])), [{self.a: 1}])
        self.assertEqual(list(self.engine.execute(self.leaf[2:])), [])
        self.assertEqual(list(self.engine.execute(self.leaf[2:3])), [])
        # Also try a non-leaf target, since that's a different code branch in
        # the iteration engine.
        b = tests.ColumnTag("b")
        target = self.leaf.with_calculated_column(b, ColumnExpression.reference(self.a))
        self.assertEqual(list(self.engine.execute(target[1:])), [{self.a: 1, b: 1}])
        self.assertEqual(list(self.engine.execute(target[:1])), [{self.a: 0, b: 0}])
        self.assertEqual(list(self.engine.execute(target[1:2])), [{self.a: 1, b: 1}])
        self.assertEqual(list(self.engine.execute(target[2:])), [])
        self.assertEqual(list(self.engine.execute(target[2:3])), [])

    def test_str(self) -> None:
        """Test str(Slice) and
        str(UnaryOperationRelation[Slice]).
        """
        relation = self.leaf[1:2]
        self.assertEqual(str(relation), f"slice[1:2]({self.leaf})")


if __name__ == "__main__":
    unittest.main()



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Utility code to aid in writing unit tests for relations and relation
algorithms.
"""

__all__ = ("ColumnTag", "RelationTestCase", "to_sql_str")

import dataclasses
import re
import unittest
from typing import Any

try:
    from sqlalchemy.dialects.sqlite.pysqlite import dialect as sql_dialect
except ImportError:
    # MyPy doesn't like this trick.
    def sql_dialect() -> Any:  # type: ignore
        """Mock sql dialect."""
        raise unittest.SkipTest("sqlalchemy SQLite dialect not available")


from ._operation_relations import BinaryOperationRelation, UnaryOperationRelation
from ._relation import Relation


def to_sql_str(sql: Any) -> str:
    """Convert a SQLAlchemy expression to a string suitable for tests.

    Parameters
    ----------
    sql
        SQLAlchemy object with a `.compile` method.

    Returns
    -------
    string : `str`
        SQL string, with all whitespace converted to single spaces.

    Notes
    -----
    This method uses the "pysqlite3" dialect, since that's the most likely one
    to be available (as it's backed by a module usually shipped with Python
    itself).

    Converting SQLAlchemy expressions to strings that are checked against
    literals is a good way to check that the `.sql` engine works as intended,
    but it's also fragile, as changes to whitespace or automatic simplification
    on the SQLAlchemy side will result in a test failure.

    Raises
    ------
    unittest.SkipTest
        Raised if `sqlalchemy` or its "pysqlite" dialect could not be imported.
    """
    return re.sub(
        r"\s+", " ", str(sql.compile(dialect=sql_dialect(), compile_kwargs={"literal_binds": True}))
    )


@dataclasses.dataclass(frozen=True)
class ColumnTag:
    """A very simple ColumnTag implementation for use in tests.

    Notes
    -----
    This class's ``__hash__`` implementation intentionally avoids the "salting"
    randomization included in the `hash` function's behavior on most built-in
    types, and it should correspond to ASCII order for single-character ASCII
    `qualified_name` values.  This means that sets of these `ColumnTag` objects
    are always deterministically ordered, and that they are *probably* sorted
    (this appears to be true, but it's really relying on CPython implementation
    details).  The former can be assumed in tests that check SQL conversions
    against expected strings (see `to_sql_str`), while the latter just makes it
    easier to write a good "expected string" on the first try.
    """

    qualified_name: str
    is_key: bool = True

    def __repr__(self) -> str:
        return self.qualified_name

    def __hash__(self) -> int:
        return int.from_bytes(self.qualified_name.encode(), byteorder="little")


def diff_relations(a: Relation, b: Relation) -> str | None:
    """Recursively compare relation trees, returning `str` representation of
    their first difference.

    Parameters
    ----------
    a : `Relation`
        Relation to compare.
    b : `Relation`
        Other relation to compare.

    Returns
    -------
    diff : `str` or `None`
        The `str` representations of the operators or relations that first
        differ when traversing the tree, formatted as '{a} != {b}'.  `None`
        if the relations are equal.
    """
    match (a, b):
        case (
            UnaryOperationRelation(operation=op_a, target=target_a),
            UnaryOperationRelation(operation=op_b, target=target_b),
        ):
            if op_a == op_b:
                return diff_relations(target_a, target_b)
            else:
                return f"{op_a} != {op_b}"
        case (
            BinaryOperationRelation(operation=op_a, lhs=lhs_a, rhs=rhs_a),
            BinaryOperationRelation(operation=op_b, lhs=lhs_b, rhs=rhs_b),
        ):
            if op_a == op_b:
                diff_lhs = diff_relations(lhs_a, lhs_b)
                diff_rhs = diff_relations(rhs_a, rhs_b)
                if diff_lhs is not None:
                    if diff_rhs is not None:
                        return f"{diff_lhs}, {diff_rhs}"
                    else:
                        return diff_lhs
                else:
                    if diff_rhs is not None:
                        return diff_rhs
                    else:
                        return None
            else:
                return f"{op_a} != {op_b}"
    if a == b:
        return None
    else:
        return f"{a} != {b}"


class RelationTestCase(unittest.TestCase):
    """An intermediate TestCase base class for relation tests."""

    def check_sql_str(self, text: str, *sql: Any) -> None:
        """Check one or more SQLAlchemy objects against the given SQL string.

        Parameters
        ----------
        text : `str`
            Expected SQL string in the pysqlite dialect, with all whitespace
            replaced by single spaces.

        *sql
            SQLAlchemy queries or column expressions to check.
        """
        for s in sql:
            self.assertEqual(to_sql_str(s), text)

    def assert_relations_equal(self, a: Relation, b: Relation) -> None:
        """Test relations for equality, reporting their difference on failure.

        Parameters
        ----------
        a : `Relation`
            Relation to compare.
        b : `Relation`
            Other relation to compare.
        """
        if diff := diff_relations(a, b):
            assert a != b, "Check that diff_relations is consistent with equality comparison."
            msg = f"{a} != {b}"
            if diff != msg:
                msg = f"{msg}:\n{diff}"
            raise self.failureException(msg)
        else:
            assert a == b, "Check that diff_relations is consistent with equality comparison."
